#ifndef _CENTROSOMES_UPDATE_POSITIONS_HPP_
#define _CENTROSOMES_UPDATE_POSITIONS_HPP_

#include "bob.h"

#include <iostream>
#include <armadillo>

#include "spatial_helpers.h"

void print_centrosomes_information(int i_spb, anchor_properties *anchors) {
    // Rigid body centrosomes
    std::cerr << "  R = {" << anchors->r_anchor[i_spb][0] << ", " 
                           << anchors->r_anchor[i_spb][1] << ", "
                           << anchors->r_anchor[i_spb][2] << "}\n";
    std::cerr << "  u = {" << anchors->u_anchor[i_spb][0] << ", " 
                           << anchors->u_anchor[i_spb][1] << ", "
                           << anchors->u_anchor[i_spb][2] << "}\n";
    std::cerr << "  v = {" << anchors->v_anchor[i_spb][0] << ", " 
                           << anchors->v_anchor[i_spb][1] << ", "
                           << anchors->v_anchor[i_spb][2] << "}\n";
    std::cerr << "  w = {" << anchors->w_anchor[i_spb][0] << ", " 
                           << anchors->w_anchor[i_spb][1] << ", "
                           << anchors->w_anchor[i_spb][2] << "}\n";
    std::cerr << "  A =\n" << anchors->A_[i_spb];
    std::cerr << "  q =\n" << anchors->q_[i_spb].as_row();
    std::cerr << "  qnorm = " << arma::dot(anchors->q_[i_spb], anchors->q_[i_spb]) << std::endl;
    std::cerr << "  translation mobility=\n" << anchors->mu_tb_[i_spb];
    std::cerr << "  rotation    mobility=\n" << anchors->mu_rb_[i_spb];
    std::cerr << "  force = {" << anchors->f_anchor[i_spb][0] << ", " 
                               << anchors->f_anchor[i_spb][1] << ", "
                               << anchors->f_anchor[i_spb][2] << "}\n";
    std::cerr << "  torque = {" << anchors->tau_local[i_spb][0] << ", " 
                                << anchors->tau_local[i_spb][1] << ", "
                                << anchors->tau_local[i_spb][2] << "}\n";
}
/*! \brief Compute lagrange correction for rigid bodies
 *
 */
double compute_lagrange_correction(const arma::vec4& q, const arma::vec4& qtilde) {
    double qqtilde = arma::dot(q, qtilde);
    double qtilde2 = arma::dot(qtilde, qtilde);

    // Solve the quadratic formulat for what is correct. Doesn't matter which
    // solution we take, as the rotations are equivalent
    double d = (2.0 * qqtilde)*(2.0 * qqtilde) - 4.0 * (qtilde2 - 1.0);
    double lambdaq = (-2.0 * qqtilde + std::sqrt(d)) / 2.0;
    return lambdaq;
}

/*! \brief Update positions of centrosomes according to rigid body dynamics
 *
 * Forces and torques are assumed to be fully 3D versions. Conversions from
 * double arrays to armadillo vec3 objects happens in this update to
 * be able to use fancy linear albegra routines
 */
void centrosomes_update_positions(system_parameters *parameters,
                                  system_properties *properties) {
    // Constants, etc
    const arma::vec3 nx = {1.0, 0.0, 0.0};
    const arma::vec3 ny = {0.0, 1.0, 0.0};
    const arma::vec3 nz = {0.0, 0.0, 1.0};
    // Bail early if SPBs are not enabled (for whatever reason)
    if (parameters->use_spb_flag != 1) {
        return;
    }

    // Set up pointers to the various matrices we need
    arma::vec4 *q = properties->anchors.q_;
    arma::mat33 *A = properties->anchors.A_;
    arma::mat33 *mu_tb = properties->anchors.mu_tb_;
    arma::mat33 *mu_rb = properties->anchors.mu_rb_;
    arma::mat33 *sqrt_mu_tb = properties->anchors.sqrt_mu_tb_;
    arma::mat33 *sqrt_mu_rb = properties->anchors.sqrt_mu_rb_;
    arma::mat::fixed<4,3> *bsub = properties->anchors.bsub_;
    const double delta_t = parameters->delta;
    const double kT = parameters->temp;
    const double thermal_diff_scale = parameters->thermal_diff_scale;
    double **r_anchor = properties->anchors.r_anchor;
    double **u_anchor = properties->anchors.u_anchor;
    double **v_anchor = properties->anchors.v_anchor;
    double **w_anchor = properties->anchors.w_anchor;

    for (int idx = 0; idx < properties->anchors.n_anchors; ++idx) {
        //std::cout << "Initial[" << idx << "]\n";
        //print_centrosomes_information(idx, &properties->anchors);
        // Set up armadillo versions of all of our information of interest
        arma::vec3 r = {properties->anchors.r_anchor[idx][0], properties->anchors.r_anchor[idx][1], properties->anchors.r_anchor[idx][2]};
        arma::vec3 u = {properties->anchors.u_anchor[idx][0], properties->anchors.u_anchor[idx][1], properties->anchors.u_anchor[idx][2]};
        arma::vec3 v = {properties->anchors.v_anchor[idx][0], properties->anchors.v_anchor[idx][1], properties->anchors.v_anchor[idx][2]};
        arma::vec3 w = {properties->anchors.w_anchor[idx][0], properties->anchors.w_anchor[idx][1], properties->anchors.w_anchor[idx][2]};

        arma::vec3 force  = {properties->anchors.f_anchor[idx][0], properties->anchors.f_anchor[idx][1], properties->anchors.f_anchor[idx][2]};
        arma::vec3 torque = {properties->anchors.tau_local[idx][0], properties->anchors.tau_local[idx][1], properties->anchors.tau_local[idx][2]};

        //// Figure out where things are going wrong
        //if (properties->i_current_step >= 33822700) {
        //    std::cerr << "Current step: " << properties->i_current_step << std::endl;
        //    print_centrosomes_information(idx, &properties->anchors);
        //}

        // Deterministic translation
        arma::vec3 dr = (((A[idx] * mu_tb[idx]) * A[idx].t()) * force) * delta_t;

        // Random thermal translation motion
        arma::vec3 theta = {
            thermal_diff_scale * gsl_ran_gaussian_ziggurat(properties->rng.r, 1.0),
            thermal_diff_scale * gsl_ran_gaussian_ziggurat(properties->rng.r, 1.0),
            thermal_diff_scale * gsl_ran_gaussian_ziggurat(properties->rng.r, 1.0)
        };
        dr += ((A[idx] * sqrt_mu_tb[idx]) * theta) * std::sqrt(2.0 * kT * delta_t);

        // Now handle the rotation part
        bsub[idx](0,0) = -q[idx][1];
        bsub[idx](0,1) = -q[idx][2];
        bsub[idx](0,2) = -q[idx][3];

        bsub[idx](1,0) =  q[idx][0];
        bsub[idx](1,1) = -q[idx][3];
        bsub[idx](1,2) =  q[idx][2];

        bsub[idx](2,0) =  q[idx][3];
        bsub[idx](2,1) =  q[idx][0];
        bsub[idx](2,2) = -q[idx][1];

        bsub[idx](3,0) = -q[idx][2];
        bsub[idx](3,1) =  q[idx][1];
        bsub[idx](3,2) =  q[idx][0];

        // Deterministic rotation part
        arma::vec4 dq = (((bsub[idx] * mu_rb[idx]) * A[idx].t()) * torque) * delta_t;

        // Random reorientation
        arma::vec3 theta_rotation = {
            thermal_diff_scale * gsl_ran_gaussian_ziggurat(properties->rng.r, 1.0),
            thermal_diff_scale * gsl_ran_gaussian_ziggurat(properties->rng.r, 1.0),
            thermal_diff_scale * gsl_ran_gaussian_ziggurat(properties->rng.r, 1.0)
        };
        dq += ((bsub[idx] * sqrt_mu_rb[idx]) * theta_rotation) * std::sqrt(2.0 * kT * delta_t);

        // Compute the lagrange correction
        arma::vec4 qtilde = q[idx] + dq;
        // Compute the corrected quaternion
        double lambdaq = compute_lagrange_correction(q[idx], qtilde);
        if (std::isnan(lambdaq)) {
            std::cerr << " quaternion correction error: " << lambdaq << std::endl;
        }

        // Actually update the information
        r += dr;
        q[idx] = qtilde + lambdaq*q[idx];

        // Update the rotation matrix and orientation vectors
        rotation_matrix_from_quaternion(A[idx], q[idx]);
        u = A[idx] * nx;
        v = A[idx] * ny;
        w = A[idx] * nz;

        // Write back into the anchor structure the proper double variables
        for (int i = 0; i < 3; ++i) {
            properties->anchors.r_anchor[idx][i] = r(i);
            properties->anchors.u_anchor[idx][i] = u(i);
            properties->anchors.v_anchor[idx][i] = v(i);
            properties->anchors.w_anchor[idx][i] = w(i);
        }
        //std::cout << "Final[" << idx << "]\n";
        //print_centrosomes_information(idx, &properties->anchors);

        // Now we have to update the positions of the MTs on the anchor
        for (al_list::iterator p = properties->anchors.anchor_list[idx].begin();
             p < properties->anchors.anchor_list[idx].end();
             p++) {
            double factor_u = dot_product(3, p->pos_rel, u_anchor[idx]);
            double factor_v = dot_product(3, p->pos_rel, v_anchor[idx]);
            double factor_w = dot_product(3, p->pos_rel, w_anchor[idx]);

            // Calculate new lab frame coordinate
            for (int i = 0; i < 3; ++i) {
                p->pos[i] = r_anchor[idx][i] + factor_u * u_anchor[idx][i] +
                    factor_v * v_anchor[idx][i] + factor_w * w_anchor[idx][i];
            }
            // Make sure that we properly attack the orientation of the anchor list
            // Orientations!
            double original_u[3] = {0.0};
            for (int i = 0; i < 3; ++i) {
                original_u[i] = p->u[i];
                p->u[i] = p->u_rel[0] * u_anchor[idx][i] +
                          p->u_rel[1] * v_anchor[idx][i] +
                          p->u_rel[2] * w_anchor[idx][i];
            }
            // Create a check if the position isn't working correctly
            for (int i = 0; i < 3; ++i) {
                if (std::isnan(p->pos[i])) {
                    // Print out the information of the centrosome and exit
                    std::cerr << "Encountered an error in SPB update code, printing information then exiting\n";
                    std::cerr << "  step: " << properties->i_current_step << std::endl;
                    print_centrosomes_information(idx, &(properties->anchors));
                    std::cerr << "  Force = " << force.as_row();
                    std::cerr << "  Torque = " << torque.as_row();
                    exit(1);
                }
            }
        }
    }
}

#endif
