#include "bob.h"

// Forward delcaration of the non-monotonic wall force stuff
double calc_non_monotonic_wall_force(double ne_ratio, double f0, double delta_r);

// SPB confinement potential that keeps SPBs locked onto the 'surface'
// of the membrane. There is a harmonic in/out force, and a harmonic
// reorientation force thatkeeps the SPB at a certain distance and 
// orientation to the NE
double spb_confinement_potential_harmonic(system_parameters *parameters,
                                          system_properties *properties,
                                          double **f_bond, double **virial,
                                          double **t_bond, int *calc_matrix) {
    double u = 0.0;

    // Bail early if the SPBs are not moving yet!
    if (parameters->use_spb_flag != 1) {
        return u;
    }

    // Loop over all SPBs in the system
    for (int idx = 0; idx < properties->anchors.n_anchors; ++idx) {
        // Construct the amount (and direction) the SPB is outside the radius
        double sys_radius = 0.5 * properties->unit_cell.h[0][0];
        // If outside the nucleus, negative
        double rmag = 0.0;
        for (int i = 0; i < 3; ++i) {
            rmag += SQR(properties->anchors.r_anchor[idx][i]);
        }
        rmag = std::sqrt(rmag);
        double rhat[3] = {0.0};
        for (int i = 0; i < 3; ++i) {
            rhat[i] = properties->anchors.r_anchor[idx][i] / rmag;
        }

        // Construct the force vector
        double forcevec[3] = {0.0};

        // If we are using a harmonic force
        if (properties->anchors.spb_confinement_type == 0) {
            double k = properties->anchors.centrosome_confinement_radial_k_;

            // Guard against pathological behavior by capping the force at some level
            double factor = -k * (rmag - sys_radius);
            double fcutoff = 0.1 / parameters->delta / MAX(properties->anchors.mu_tb_[idx](0,0),
                                                           properties->anchors.mu_tb_[idx](1,1));
            if (factor < 0.0) {
                if (-1.0 * factor > fcutoff) {
                    factor = -1.0 * fcutoff; 
                    std::cerr << " *** Force exceeded fcutoff spb_confinement_potential_harmonic_bd radial ***\n";
                }
            } else {
                if (factor > fcutoff) {
                    factor = fcutoff;
                    std::cerr << " *** Force exceeded fcutoff spb_confinement_potential_harmonic_bd radial ***\n";
                }
            }

            for (int i = 0; i < 3; ++i) {
                forcevec[i] = factor * rhat[i];
            }
        } else if (properties->anchors.spb_confinement_type == 1) {
            double f0 = properties->anchors.centrosome_confinement_f0_;

            double delta_r = ABS(sys_radius - rmag);
            double factor = calc_non_monotonic_wall_force(parameters->ne_ratio,
                                                          f0,
                                                          delta_r);

            // Check the sign of the force, want to move outwards if we are in the nucleoplasm
            if (rmag > sys_radius) {
                for (int i = 0; i < 3; ++i) {
                    forcevec[i] = -1.0 * factor * rhat[i];
                }
            } else {
                for (int i = 0; i < 3; ++i) {
                    forcevec[i] = 1.0 * factor * rhat[i];
                }
            }
        }

        ////Based on the potential, calculate the torques on the system
        //double uhat_dot_rhat = dot_product(3, rhat, properties->anchors.u_anchor[idx]);
        //double uhat_cross_rhat[3] = {0.0};
        //cross_product(properties->anchors.u_anchor[idx], rhat, uhat_cross_rhat, 3);
        //double kr = properties->anchors.centrosome_confinement_angular_k_;
        //// The torque is thusly
        //double torquevec[3] = {0.0};

        //for (int i = 0; i < 3; ++i) {
        //    torquevec[i] = -kr * (uhat_dot_rhat + 1.0) * uhat_cross_rhat[i];
        //}

        //// Now calculate the torque
        //// Really, this wants to align the uhat vector with the inward facing rhat vector
        //double rhati[3] = {-1.0*rhat[0], -1.0*rhat[1], -1.0*rhat[2]};
        //double rhati_dot_uhat = dot_product(3, rhati, properties->anchors.u_anchor[idx]); 
        //double rhati_cross_uhat[3] = {0.0};
        //cross_product(rhati, properties->anchors.u_anchor[idx], rhati_cross_uhat, 3);
        //double kr = properties->anchors.centrosome_confinement_angular_k_;
        //// The torque is simple to calculate to face inwards
        //double torquevec[3] = {0.0};
        //for (int i = 0; i < 3; ++i) {
        //    torquevec[i] = kr * (rhati_dot_uhat - 1.0) * rhati_cross_uhat[i];
        //}

        //if (properties->i_current_step % 100000 == 0) {
        //    std::cout << "--------\n";
        //    std::cout << "SPB confinement potential, step: " << properties->i_current_step << ", spb: " << idx << std::endl;
        //    std::cout << "  rmag = " << rmag << std::endl;
        //    std::cout << "  forcevec  = (" << forcevec[0] << ", " << forcevec[1] << ", " << forcevec[2] << ")\n";
        //    std::cout << "  forcemag  = " << std::sqrt(forcevec[0]*forcevec[0] + forcevec[1]*forcevec[1] + forcevec[2]*forcevec[2]) << std::endl;
        //    std::cout << "  torquevec = (" << torquevec[0] << ", " << torquevec[1] << ", " << torquevec[2] << ")\n";
        //    std::cout << "  torquemag = " << std::sqrt(torquevec[0]*torquevec[0] + torquevec[1]*torquevec[1] + torquevec[2]*torquevec[2]) << std::endl;
        //}
        
        // Add the contribution to the total forces
        for (int i = 0; i  < 3; ++i) {
            properties->anchors.f_anchor[idx][i] += forcevec[i];
            properties->anchors.tau_local[idx][i] += torquevec[i];
        }
    }

    return u;
}
